.g-oauth-container
  margin-top 20px

  .g-oauth-section-header
    height 2px
    background-color #dcd6d6
    margin-bottom 16px
    font-size 12px

  .g-oauth-section-header-text
    display inline
    background-color #fff
    position relative
    top -7px
    left 39%
    padding 0 7px
    color #878080

.g-oauth-button
  border-radius 3px
  padding 7px 11px
  display inline-block
  cursor pointer
  margin-right 18px

  .g-oauth-button-icon
    display inline-block
    padding-right 6px
    border-right 1px solid white
  .g-oauth-button-text
    display inline-block
    padding 0 20px

.g-oauth-button-google
  background-color #d84936
  border 1px solid #ac3d2f
  color white

  &:hover
    background-color #c83926

  .g-oauth-button-icon
    border-right-color #ac3d2f
