.g-oauth-provider-list-title
  margin 10px 0
  font-size 16px
  color #555
  border-bottom 1px solid #eee

.g-oauth-value-container
  margin 10px 0 10px 0
  padding 3px 9px
  border 1px solid #ddd
  background-color #f6f6f6

  .g-oauth-value
    display inline-block
    margin-left 15px
    font-family monospace
